#include <joyos.h>

#define HEADING_SCALE (3)
#define HEADING_CORRECTION_LIMIT (45)
#define TRANSLATION_SCALE (0.2)

float PIDCorrectionVal;			//total correction
extern uint8_t forwardVel;		//externally defined forward bias, mostly for testing
float headingCorrectionVal;		//motor correction value as judged by heading PID
float translationCorrectionVal;	//motor bias value as judged by distance PID

//outputs velocity as a correctly typed and ranged value
int16_t limitVel(float vel){
	//	printf("limitvel running: %f\n", vel);  //for debugging
	if (vel < -255) return -255;
	if (vel > 255) return 255;				//prevents sending the motors an unacceptable value
	return (int16_t)vel;
}

void setHeading(float correctionAmount){
	//  motor_set_vel(0, -limitVel(forwardVel - correctionAmount*.5));
	//  motor_set_vel(1, -limitVel(forwardVel + correctionAmount*.5));  THIS IS A REMNANT OF OLD CODE ONLY FOR REFERENCE

	//should reduce kickback after long delays being stuck
	if (correctionAmount > HEADING_CORRECTION_LIMIT){
		correctionAmount = HEADING_CORRECTION_LIMIT;
	}


	if(correctionAmount<-HEADING_CORRECTION_LIMIT){
		correctionAmount = -HEADING_CORRECTION_LIMIT;
	}

	//end kickback reduction

	headingCorrectionVal = correctionAmount*HEADING_SCALE;  //scales heading correction by constant since heading is always important

}

//sends error from error from translationPID 
//to be scaled and reduced by the heading error
//allowing the robot to pivot if the error is greater enough
void setTranslation(float correctionAmount){
	translationCorrectionVal = (correctionAmount*TRANSLATION_SCALE) - (fabs(headingCorrectionVal))*HeadingSubtract;

	//it is so it doesn't loop around the target because it can't turn fast enough
	if (translationCorrectionVal <=0){
		translationCorrectionVal = 0;			//prevents error from being negative
	}
}


float getTranslationErr(void){		//sets forwards or backwards
	if(dir.forward){
		return -translationCorrectionVal;
	}else{
		return translationCorrectionVal;
	}
}

float getHeadingErr(void){			//sets left or right
	if(dir.forward){
		if(dir.left){
			return -headingCorrectionVal;
		}else{
			return headingCorrectionVal;
		}
	}else{
		if(dir.left){
			return headingCorrectionVal;
		}else{
			return -headingCorrectionVal;

		}
	}
}