#include <joyos.h>

//converts a value from one range to the equivalent in another
float map(float x, float in_min, float in_max, float out_min, float out_max)
{
	return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

//allows the goal to be changed by passing x and y coordinates and the heading of the target
void update_goal(long int x, long int y){  
	goal.x = x;
	goal.y = y;

	goal.theta = goal_heading();		//update angle to goal using vision
	//goal.theta = 0;
}


//fed into the velocity PID for estimating distance to target
// ***should be distance from current point to goal
float line_dist(void){

	//Commented for testing 

	double d2 = (me.x - goal.x) * (me.x - goal.x) + (me.y - goal.y) * (me.y - goal.y);
	float dist = (float) sqrt(d2);					//returns the straight line distance between the robots last updated position and the goal position
	return dist;

	//use the frob instead
	//return (float) frob_read_range(0, 1500);

}

//convert from radians to degrees
float to_degrees(float rad){   
	return 180.*rad/M_PI;
}

//returns the angle heading from current location to goal point in degrees form -179 to 180
float goal_heading(void){
	if (goal.x == me.x){
		return 0;
	}

	//goalHeading = radians for now
	float goalHeading = (float) atan2((double) (goal.y-me.y), (double) (goal.x-me.x));  //calculates the goal direction as compared to the positive x axis

	//-180 would be an invalid valid for output range, use equivalent 180 instead
	if (goalHeading == -M_PI){
		return 180;
	}
	return to_degrees(goalHeading);
}


void update_motors_capture(void){
	float headErr = getHeadingErr();

	float smallHeadError = fabs(getHeadingErr());

	float mapHeadError = map((headErr), MIN_ERROR_HEADING_CAPTURE, SPEED_CHANGE_HEADING_CAPTURE, 30, 45);
	float negMapHeadError = map((headErr), -MIN_ERROR_HEADING_CAPTURE, -SPEED_CHANGE_HEADING_CAPTURE, -30, -45);


	if (smallHeadError > SPEED_CHANGE_HEADING_CAPTURE){
		motor_set_vel(0, limitVel(+ headErr));  //large errors, corrects largely
		motor_set_vel(1, limitVel(- headErr));
	}else if(smallHeadError <= SPEED_CHANGE_HEADING_CAPTURE && headErr >= MIN_ERROR_HEADING_CAPTURE){
		motor_set_vel(0, limitVel(+ mapHeadError));  //small errors are still significant, AND in the right direction
		motor_set_vel(1, limitVel(- mapHeadError));
	}else if(smallHeadError <= SPEED_CHANGE_HEADING_CAPTURE && headErr <= -MIN_ERROR_HEADING_CAPTURE){
		motor_set_vel(0, limitVel(+ negMapHeadError));  //small errors are still significant, AND in the right direction
		motor_set_vel(1, limitVel(- negMapHeadError));
	}else {
		motor_set_vel(0, 0);  //with in allowable error
		motor_set_vel(1, 0);
	}
}

void recalibrate(void){
		pause(100);
		update_position();
		gyro_set_degrees(map((float)((long int) game.coords[0].theta),-2048.,2047.,-179.,180.));
		pause(5);
}


void update_motors_heading(void){
	float headErr = getHeadingErr();

	float smallHeadError = fabs(getHeadingErr());

	float mapHeadError = map((headErr), MIN_ERROR_HEADING_CAPTURE, SPEED_CHANGE_HEADING, MIN_HEADING_MAP, SPEED_CHANGE_HEADING);
	float negMapHeadError = map((headErr), -MIN_ERROR_HEADING_CAPTURE, -SPEED_CHANGE_HEADING, -MIN_HEADING_MAP, -SPEED_CHANGE_HEADING);


	if (smallHeadError > SPEED_CHANGE_HEADING){
		motor_set_vel(0, limitVel(+ headErr));  //large errors, corrects largely
		motor_set_vel(1, limitVel(- headErr));
	}else if(smallHeadError <= SPEED_CHANGE_HEADING && headErr >= MIN_ERROR_HEADING_CAPTURE){
		motor_set_vel(0, limitVel(+ mapHeadError));  //small errors are still significant, AND in the right direction
		motor_set_vel(1, limitVel(- mapHeadError));
	}else if(smallHeadError <= SPEED_CHANGE_HEADING && headErr <= -MIN_ERROR_HEADING_CAPTURE){
		motor_set_vel(0, limitVel(+ negMapHeadError));  //small errors are still significant, AND in the right direction
		motor_set_vel(1, limitVel(- negMapHeadError));
	}else {
		motor_set_vel(0, 0);  //with in allowable error
		motor_set_vel(1, 0);
	}
}


//sends the data through to the motors.  Is this the proper way to do this?  maybe... ;)
void update_motors(void){	

	float transErr = getTranslationErr();   // bias forwards, real deal
	//float transErr = 0;					// don't bias forwards, we are testing!!
	float headErr = getHeadingErr();
	float smallHeadError = fabs(getHeadingErr());
	float smallTransError = fabs(getTranslationErr());

	float mapHeadError = map((headErr), MIN_ERROR_HEADING, SPEED_CHANGE_HEADING, MIN_HEADING_MAP, SPEED_CHANGE_HEADING);
	float negMapHeadError = map((headErr), -MIN_ERROR_HEADING, -SPEED_CHANGE_HEADING, -MIN_HEADING_MAP, -SPEED_CHANGE_HEADING);


	if (smallTransError>=SPEED_CHANGE){
		if (smallHeadError > SPEED_CHANGE_HEADING){
			motor_set_vel(0, limitVel(transErr + headErr));  //large errors, corrects largely
			motor_set_vel(1, limitVel(transErr - headErr));
		}else if(smallHeadError <= SPEED_CHANGE_HEADING && headErr >= MIN_ERROR_HEADING){
			motor_set_vel(0, limitVel(transErr + mapHeadError));  //small errors are still significant, AND in the right direction
			motor_set_vel(1, limitVel(transErr - mapHeadError));
		}else if(smallHeadError <= SPEED_CHANGE_HEADING && headErr <= -MIN_ERROR_HEADING){
			motor_set_vel(0, limitVel(transErr + negMapHeadError));  //small errors are still significant, AND in the right direction
			motor_set_vel(1, limitVel(transErr - negMapHeadError));
		}else{
			motor_set_vel(0, limitVel(transErr));  //too small errors are ignored, this does no small value correction to the translation
			motor_set_vel(1, limitVel(transErr));
		}
	}else if(smallTransError < SPEED_CHANGE && smallTransError >= MIN_ERROR_TRANSLATION){
		if (transErr>=0){
			transErr = map((transErr), MIN_ERROR_TRANSLATION , SPEED_CHANGE, 70, SPEED_CHANGE);
			//printf("Motors: %f  Error: %f\n", transErr, smallTransError);
		}else{
			transErr = -map((-transErr), MIN_ERROR_TRANSLATION , SPEED_CHANGE, 70, SPEED_CHANGE);
		}

		if (smallHeadError > SPEED_CHANGE_HEADING){
			motor_set_vel(0, limitVel(transErr + headErr));  //large errors, corrects largely
			motor_set_vel(1, limitVel(transErr - headErr));
		}else if(smallHeadError <= SPEED_CHANGE_HEADING && headErr >= MIN_ERROR_HEADING){
			motor_set_vel(0, limitVel(transErr + mapHeadError));  //small errors are still significant, AND in the right direction
			motor_set_vel(1, limitVel(transErr - mapHeadError));
		}else if(smallHeadError <= SPEED_CHANGE_HEADING && headErr <= -MIN_ERROR_HEADING){
			motor_set_vel(0, limitVel(transErr + negMapHeadError));  //small errors are still significant, AND in the right direction
			motor_set_vel(1, limitVel(transErr - negMapHeadError));
		}else{
			motor_set_vel(0, limitVel(transErr));  //too small errors are ignored, this does no small value correction to the translation
			motor_set_vel(1, limitVel(transErr));
		}


	}else{
		if (smallHeadError > SPEED_CHANGE_HEADING){
			motor_set_vel(0, limitVel( + headErr));  //large errors, corrects largely
			motor_set_vel(1, limitVel( - headErr));
		}else if(smallHeadError <= SPEED_CHANGE_HEADING && headErr >= MIN_ERROR_HEADING){
			motor_set_vel(0, limitVel( + mapHeadError));  //small errors are still significant, AND in the right direction
			motor_set_vel(1, limitVel( - mapHeadError));
		}else if(smallHeadError <= SPEED_CHANGE_HEADING && headErr <= -MIN_ERROR_HEADING){
			motor_set_vel(0, limitVel( + negMapHeadError));  //small errors are still significant, AND in the right direction
			motor_set_vel(1, limitVel( - negMapHeadError));
		}else{
			motor_set_vel(0, 0);  //small errors are still significant, AND in the right direction
			motor_set_vel(1, 0);
		}
	}

}

void update_all(long int x, long int y){
	update_position();
	update_goal(x, y);
	update_pid(&headingPID);
	update_pid(&velPID);
}

//this will need to get a LOT more advanced
//using previous values and other current information,
//good enough for now though
void update_position(void){ 

	//get data from vision control system

	copy_objects();

	//update info based on vision data

	me.x = (long) game.coords[0].x;
	me.y = (long) game.coords[0].y;

	//uint32_t currTime = position_microtime[0];
	//if ( lastTime != currTime){
	//	gyro_set_degrees(map((float)((long int) game.coords[0].theta),-2048.,2047.,-179.,180.));  //calibrate gyro on incoming data
	//}

	//me.theta = map((float)((long int) game.coords[0].theta),-2048.,2047.,-179.,180.);	//olddeal
	me.theta = gyro_get_degrees();												//real deal, makes use of gyro correction


	//continuously update goal point based on current location
	//update_goal(goal.x,goal.y);					//testing

	//update_goal(game.coords[2].x, game.coords[2].y);	//realdeal

	//headingPID.goal = goal.theta;   // needs to be 0!!!
	//lastTime = currTime;
}

//gets current heading
float get_heading(void){

	return me.theta;
}




//////////////////////////////-------------------------------------------------------BEGIN JAKE'S FAULT-----------------------------------------------


//takes in angle in -180 to 180 DEGREE range (0 degrees at positive x)
//and outputs the corresponding angle 0 to 360 DEGREES
float convertToPositiveAngle(float a) {

	if ( a < 0 ) {
		a = 360.+a;
	}
	return fmod((double) a, 360.);

}

//given angles a1 and a2 (in DEGREES!)
//returns smallest angle in degrees between them
float getShortestAngle(float a1,float a2) {

	float a = fmod((double) fabs(a1-a2), 360.);
	if ( a > 180 ) {
		a = 360-a;
	}
	return a;

}




// given a current heading and a goal heading (floats in DEGREES -180 to 180)
// returns the shortest turn angle to reach that heading
// - a forwards/backwards turn
// - a right/left turn
// - angle of turn from forwards/backwards in right/left direction of turn in DEGREES 0 to 90
//
// modifies dir struct 
// dir.forward:		forward = TRUE		backwards = FALSE
// dir.left:		left = TRUE		right = FALSE
// dir.theta: 		0 - 90 degrees
float getBestAngleTurn() {

	float _goalHeading = goal.theta;
	float _currentHeading = me.theta;


	float pHeading = convertToPositiveAngle(_currentHeading);
	float pGoalAngle = convertToPositiveAngle(_goalHeading);

	float shortAngle = getShortestAngle(pHeading,pGoalAngle);
	float shortAngleLeft = getShortestAngle(pHeading+90,pGoalAngle);

	if ( shortAngle <= 90 ) { //FORWARD
		if ( shortAngleLeft <= 90 ) { //LEFT
			//forward, left, shortAngle
			dir.forward = true;
			dir.left = true;
			dir.theta = shortAngle;
		}
		else { //RIGHT
			//forward, right, shortAngle
			dir.forward = true;
			dir.left = false;
			dir.theta = shortAngle;
		}
	}
	else { //BACKWARD
		if ( shortAngleLeft <= 90 ) { //LEFT
			//backward, left, 180-shortAngle
			dir.forward = false;
			dir.left = true;
			dir.theta = 180-shortAngle;
		}
		else { //RIGHT
			//backward, right, 180-shortAngle
			dir.forward = false;
			dir.left = false;
			dir.theta = 180-shortAngle;
		}
	}

	return dir.theta; //value to PID

}


float heading_select(){

	if (!headingSelect){
		return getBestAngleTurn();
		}else{

		float meThetaConverted = convertToPositiveAngle(me.theta);
		float goalThetaConverted = convertToPositiveAngle(goal.theta);
		float shortAngle = getShortestAngle(meThetaConverted,goalThetaConverted);
		float shortAngleLeft = getShortestAngle(meThetaConverted+90,goalThetaConverted);
		if ( shortAngleLeft<=90 ) {
			dir.left = 0;
		}
		else {
			dir.left = 1;
		}
		dir.theta = shortAngle;
		return dir.theta;
	}
	}
