/*
* The MIT License
*
* Copyright (c) 2007 MIT 6.270 Robotics Competition
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
*
*/


/* ----------------------------HARDWARE SETUP----------------------------------------------------
gyro = pin 11
motor 0 on motor pin 0, red wire in middle of header
motor 1 on motor pin 1, red wire in middle of header
capture motor is on motor pin 3 same as above
arm servo is servo 0
lever servo is servo 1
gate servo is servo 2
lever switch is on digital 6
dump switch is on digital 7
*/




#include <joyos.h>

extern volatile uint8_t robot_id;  //Vision stuff, don't delete


//make sure libraries are on top, at least for now  O.o


uint32_t lastTime;

uint32_t matchStartTime; //used keep track of how much time is left in the match

int usetup (void) {

	for(int i = 0; i<6; i++){
		digital_write(i, 0);
	}


	lastTime = 0;
	goal.x = 0;
	goal.y = 0;
	newCommand = 0;
	HeadingSubtract = 1.25;
	robot_id = 8;  //Vision stuff, don't delete
	headingSelect = false;

	currentCommand.type = IDLE;

	init_controllers();
	//gyro_set_degrees(0);

	init_lock(&current_command_lock, "currentCommand lock"); //initialize the struct

	init_nodes();
	
	init_stuck_moving();

	//led indicators
	for(int i = 0; i<2; i++){
		digital_write(i, 1);
		pause(80);
	}
	for(int i = 0; i<4; i++){
		digital_write(i+2, 1);
		digital_write(i, 0);
		pause(80);
	}
	for(int i = 4; i<6; i++){
		digital_write(i, 0);
		pause(80);
	}
	for(int i = 5; i>3; i--){
		digital_write(i, 1);
		pause(80);
	}
	for(int i = 5; i>1; i--){
		digital_write(i-2, 1);
		digital_write(i, 0);
		pause(80);
	}
	for(int i = 1; i>=0; i--){
		digital_write(i, 0);
		pause(80);
	}

	//printf("ready to go!/n");

	return 0;
}

int umain (void) {

	matchStartTime = get_time(); //used keep track of how much time is left in the match

	//printf("running umain\n");
	update_position();
	gyro_set_degrees(map((float)((long int) game.coords[0].theta),-2048.,2047.,-179.,180.));  //zero the gyro to the current heading

	if(me.x >= 0){
		team = BLUE;
	}else{
		team = RED;
	}

	//printf("about to start strategy thread\n");
	create_thread(&strategy, STACK_DEFAULT, 0, "stategy thread");
	//printf("strategy thread started\n");

	//uint16_t loopsNumber=1;
	uint8_t lastLitCommand = 0;

		
while (1){    
		//printf("in primary while loop  number: %u\n", loopsNumber);
		headingSelect = false;
		if ( newCommand == 1 ) {
			//printf("new command\n");
			newCommand = 0;
			acquire(&current_command_lock);

			digital_write(lastLitCommand, 0);

			if(currentCommand.type == DUMP){
				digital_write(5, 1);
				lastLitCommand = 5;
			}else{
				digital_write(currentCommand.type, 1);
				lastLitCommand = currentCommand.type;
			}

			if ( currentCommand.type == IDLE ) {
				release(&current_command_lock);

				//printf("idling\n");
				motor_set_vel(0,0);
				motor_set_vel(1,0);
				motor_set_vel(2,0);
				servo_set_pos(0,435);
				servo_set_pos(1,256);

				//if (get_time() - matchStartTime > 110000){  //dump?


			}
			else if ( currentCommand.type == MOVING ) {
				release(&current_command_lock);

				if(((currentCommand.pointInfo) % N_TERRITORY_POINTS) == TERRITORY_CENTER){ 

					velPID.kp = -2.;  //goes faster for center to center
					HeadingSubtract = 1.1;
				}else if(((currentCommand.pointInfo) % N_TERRITORY_POINTS) == TERRITORY_GEARBOX){
					velPID.kp = -1.25;

				}else if(((currentCommand.pointInfo) % N_TERRITORY_POINTS) == TERRITORY_LEVER){
					HeadingSubtract = .75;
					velPID.kp = -1.25;
				}

				//printf("moving\n");
				move_to_point( currentCommand.x, currentCommand.y );
				velPID.kp = VELOCITY_GAIN_PROPORTIONAL;
				headingPID.kp = HEADING_GAIN_PROPORTIONAL;
				HeadingSubtract = 1.25;

			}else if ( currentCommand.type == TURNING ) {
				release(&current_command_lock);

				headingSelect = true;

				//printf("turning to %f\n", currentCommand.heading);
				turn_to_heading( currentCommand.heading );

				headingSelect = false;

			}else if ( currentCommand.type == FACING ) {
				release(&current_command_lock);
				headingSelect = false;

				//printf("facing\n");
				turn_to_point( currentCommand.x, currentCommand.y );

			}else if ( currentCommand.type == CAPTURE ) {
				release(&current_command_lock);

				//printf("capturing\n");
				uint32_t timeDelay = get_time();
				servo_capture();
				int j=0;
				while(!newCommand){

					if(motor_get_current_MA(2)>1200 && (get_time() - timeDelay>500)){  //pulls arm in if stalled
						servo_set_pos(0, 255+j);
						j+=2;
					}else if(motor_get_current_MA(2)<800 && (get_time() - timeDelay>500)  && (j > -25)){  //pushes arm out if freewheeling, to a limit
						servo_set_pos(0, 255+j);
						j-=2;
					}
					pause(10); 
					yield();}
				servo_return();

			}else if ( currentCommand.type == PULL ) {
				release(&current_command_lock);

				//printf("pulling\n");

				int myTerritory = point_get_territory(me.x, me.y);
				//printf("forwards to wall\n");

				motor_set_vel(0, -120);
				motor_set_vel(1, -120);

				while(digital_read(6)){
					pause(10);
					//	printf("still no button press\n");
				}

				motor_set_vel(0, 0);
				motor_set_vel(1, 0);

				uint32_t timeDelay = get_time();

				int ballsInTarget = game.territories[myTerritory].remaining;
				int k = 0;
				int ballsNow = ballsInTarget;

				while(!newCommand){  //rate limit not activated and fewer than 5 balls captured
					if(ballsNow != game.territories[myTerritory].remaining){
						//printf("got one!\n");
						k = 0;
						for(int i = 0; i<2; i++){
							digital_write(i, 1);
							pause(80);
						}
						for(int i = 0; i<4; i++){
							digital_write(i+2, 1);
							digital_write(i, 0);
							pause(80);
						}
						for(int i = 4; i<6; i++){
							digital_write(i, 0);
							pause(80);
						}
						nBallsCollected++; //number of balls in hopper increased

					}
					/*
					if(k>=3){                    //tried 3 times with no success, back up and try more
					printf("gotta back up, too close or something\n");
					motor_set_vel(0,100);
					motor_set_vel(1,100);
					printf("stopping back up\n");
					pause(100);
					motor_set_vel(0,0);
					motor_set_vel(1,0);
					k = 0;
					}
					*/
					//printf("trying to pull\n");
					ballsNow = game.territories[myTerritory].remaining;
					k++;
					copy_objects();
				}
				servo_set_pos(1, 256);
				//printf("must be done with balls!\n");

			}else if ( currentCommand.type == DUMP ) {
				release(&current_command_lock);

				motor_set_vel(0, 120);
				motor_set_vel(1, 120);
				//printf("go to wall\n");
				while(!digital_read(7)){
					pause(10);
					//	printf("still no button press\n");
				}
				motor_set_vel(0, 0);
				motor_set_vel(1, 0);
				servo_set_pos(2, 256);
				pause(3000);  //wait for gate to let balls out
				servo_set_pos(2, 75);
				dumping = false;   ///let strategy know we are done
				yield();

			}else{
				for(int i = 0; i<6; i++){
					digital_write(i, 1);
				}
				release(&current_command_lock);
			}

		}
		//printf("exiting new command sequence and yielding\n");
		//loopsNumber++;
		yield();
	}

	return 0;

}

int hardware_test(void){
			////////////////////////////////////////////--------------WARM-UP CODE!!!!!!!!!!!---------------////////////////
	for(int i=0; i<3; i++){
	servo_set_pos(2, 256);
	pause(300);
servo_set_pos(2, 75);
pause(300);
	}

	servo_set_pos(0, 255);
	motor_set_vel(2, 255);
	pause(1000);
		motor_set_vel(2, 0);
	servo_set_pos(0, 435);

	motor_set_vel(0, -150);
	motor_set_vel(1, 150);
	pause (500);
	motor_set_vel(0, 150);
	motor_set_vel(1, -150);
	pause(500);
	motor_set_vel(0, 0);
	motor_set_vel(1, 0);


	while(get_time()< 15000){
	if(gyro_get_degrees() >0 && gyro_get_degrees() <90){
		digital_write(5, 1);
	}else if(gyro_get_degrees() >90){
		digital_write(4, 1);
	}else if(gyro_get_degrees() <0 && gyro_get_degrees() >-90){
		digital_write(3, 1);
	}else if(gyro_get_degrees() <-90 ){
		digital_write(2, 1);
	}

	digital_write(1, !digital_read(6));
	digital_write(0, digital_read(7));
	}
	return 1;
}